# flake8: noqa
from .consumer import (
    Consumer as RabbitConsumer,
    ConsumerType as RabbitConsumerType,
)
from .parameters import RabbitConnectionParameters
from .producer import Producer as RabbitProducer



# flake8: noqa
from .event import FranzEvent
from .exceptions import InvalidMessage, SerializationError
from .kafka import KafkaConsumer, KafkaProducer
from .rabbitmq import (
    RabbitConnectionParameters,
    RabbitConsumer,
    RabbitConsumerType,
    RabbitProducer,
)



import uuid

import pika

from franz import base
from .parameters import RabbitConnectionParameters


class Producer(base.BaseProducer):
    """
    A simple producer wrapper for pika/RabbitMQ.

    Examples
    --------
    >>> from franz import RabbitProducer
    >>> from myapp.models import SomeModel
    >>> instance = SomeModel.objects.get(pk=1)
    >>> with RabbitProducer(app_name='web') as p:
    >>>     p.send_message('myapp.logs.critical', instance)
    """

    PERSISTENT_DELIVERY_MODE = 2

    def __init__(self, app_name, parameters=None):
        self._parameters = parameters
        self._app_name = app_name
        self._connection = pika.BlockingConnection(self.parameters)

        self._channel = self._connection.channel()

    @property
    def parameters(self):
        if self._parameters is None:
            self._parameters = RabbitConnectionParameters()
        return self._parameters.connection_parameters

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._connection.close()

    def get_properties(self, correlation_id=None):
        return pika.BasicProperties(
            delivery_mode=self.PERSISTENT_DELIVERY_MODE,
            correlation_id=self.get_correlation_id(correlation_id),
        )

    def get_correlation_id(self, correlation_id=None):
        return '{uuid}.{app_name}'.format(
            uuid=correlation_id or str(uuid.uuid4()),
            app_name=self._app_name,
        )

    def send_message(self, topic, message, correlation_id=None):
        self._check_message_is_valid_event(message)
        self._send_message_to_topic(topic, message, correlation_id)

    def _send_message_to_topic(self, topic, message, correlation_id=None):
        """
        Send a message to RabbitMQ based on the routing key (topic).

        Parameters
        ----------
        topic : str
            The routing key (topic) where the message should be sent to.
        message : FranzEvent
            The message to be sent.

        Raises
        ------
        franz.InvalidMessage
        """
        exchange = self.get_exchange_name(topic)
        self._channel.exchange_declare(
            exchange=exchange,
            exchange_type='topic',
            durable=True,
        )
        self._channel.basic_publish(
            exchange,
            topic,
            self.serialize_message(message),
            properties=self.get_properties(correlation_id=correlation_id),
        )

    @classmethod
    def get_exchange_name(cls, topic):
        return topic.split('.')[0]



class FranzEvent:
    """
    Classes that want to be sent as messages to Kafka
    should inherit from this class.
    """

    def serialize(self):
        """
        Converts the current python object to one that bson can serialize.
        Complex objects should define their own `.serialize()` method.

        Returns
        -------
        dict
        """
        return self.__dict__.copy()



import bson

from kafka import KafkaConsumer


class Consumer:
    """
    A simple wrapper around KafkaConsumer.

    Examples
    --------
    >>> from franz import KafkaConsumer
    >>> consumer = KafkaConsumer('TopicA', 'Topic-*')
    >>> for message in consumer:
    >>>     print(message.value)
    """
    def __init__(self, *topics, server='localhost'):
        """
        Parameters
        ----------
        topics : tuple
            String arguments that are exact topic names like 'MyTopic' or
            pattern names like 'MyTopic-*'
        """
        self.topics = topics
        self.consumer = KafkaConsumer(
            value_deserializer=bson.loads,
            bootstrap_servers=server,
        )
        self.consumer.subscribe(pattern=self._make_topic_regex())

    def __iter__(self):
        return self.consumer.__iter__()

    def _make_topic_regex(self):
        """
        Makes an OR regex for all topics passed in. KafkaConsumer doesn't allow
        for patterns and regular topics by default so this allows support for
        both.

        Returns
        -------
        str
            The OR'd regex for all topics.
        """
        return '|'.join("({})".format(topic) for topic in self.topics)



from setuptools import setup


__version__ = "0.0.9"


setup(
    name="franz",
    version=__version__,
    url="https://github.com/eshares/franz",
    download_url="https://github.com/eshares/franz/tarball/{version}".format(
        version=__version__,
    ),
    author="Carta, Inc.",
    author_email="engineering@carta.com",
    description=(
        "Event broker built on top of kafka and rabbitmq; used to handle micro"
        " services message exchange."
    ),
    long_description=open('README.md').read(),
    license="MIT",
    keywords="microservices broker event kafka rabbitmq",
    install_requires=[
        "kafka-python==1.3.5",
        "bson==0.5.0",
        "pika==0.11.0",
    ],
    packages=[
        "franz",
        "franz.kafka",
        "franz.rabbitmq",
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    zip_safe=True,
)



from kafka import KafkaProducer

from franz import base


class Producer(base.BaseProducer):
    """
    A simple wrapper around kafka.KafkaProducer.

    Examples
    --------
    >>> from franz import KafkaProducer
    >>> from myapp.models import SomeModel
    >>> instance = SomeModel.objects.get(pk=1)
    >>> p = KafkaProducer()
    >>> p.send_message('test', instance)
    """

    def __init__(self, host='localhost', port=None):
        self.host = host
        self.port = port

        self.producer = KafkaProducer(
            bootstrap_servers=self.connection_string,
            value_serializer=Producer.serialize_message,
        )

    @property
    def connection_string(self):
        return '{host}{port}'.format(
            host=self.host,
            port=':{}'.format(self.port) if self.port else '',
        )

    def _send_message_to_topic(self, topic, message):
        """
        Send a message to a Kafka topic.

        Parameters
        ----------
        topic : str
            The kafka topic where the message should be sent to.
        message : FranzEvent
            The message to be sent.

        Raises
        ------
        franz.InvalidMessage
        """
        message_result = self.producer.send(topic, message)
        self.check_for_message_exception(message_result)
        return message_result

    @classmethod
    def check_for_message_exception(cls, message_result):
        """
        Makes sure there isn't an error when sending the message.
        Kafka will silently catch exceptions and not bubble them up.

        Parameters
        ----------
        message_result : FutureRecordMetadata
        """
        exception = message_result.exception

        if isinstance(exception, BaseException):
            raise exception



import logging

import bson
import pika

from .parameters import RabbitConnectionParameters
from .producer import Producer


class ConsumerType:
    # User constants
    SINGLE_PROCESS = 'single'  # Single consumer
    MULTIPLE_FAST_PROCESSES = 'multiple-fast'  # Multiple fast consumers
    MULTIPLE_SLOW_PROCESSES = 'multiple-slow'  # Multiple slow consumers

    # Prefetch values
    _SINGLE_FAST_CONSUMER = 0
    _MULTIPLE_FAST_CONSUMERS = 20
    _MULTIPLE_SLOW_CONSUMERS = 1

    SPEED_TO_PREFETCH = {
        SINGLE_PROCESS: _SINGLE_FAST_CONSUMER,
        MULTIPLE_FAST_PROCESSES: _MULTIPLE_FAST_CONSUMERS,
        MULTIPLE_SLOW_PROCESSES: _MULTIPLE_SLOW_CONSUMERS,
    }


class Consumer:
    """
    A simple consumer wrapper for pika/RabbitMQ.

    Examples
    --------
    >>> from franz import RabbitConsumer
    >>> def callback(correlation_id, topic, body):
    >>>     msg = '[+] Received message {} from {}. ID: {}'.format(
    >>>         body, topic, correlation_id)
    >>>     print(msg)
    >>> with RabbitConsumer('myapp.logs.critical', queue='logs') as consumer:
    >>>     consumer.consume_messages(callback)
    """

    def __init__(self, *topics, queue='', parameters=None,
                 consumer_type=ConsumerType.SINGLE_PROCESS):
        """
        Parameters
        ----------
        topics : tuple
            The topics to consume from.
        queue : str
            The RabbitMQ queue to consume messages from.
        parameters : RabbitConnectionParameters
            The parameters to connect to the RabbitMQ server.
        consumer_type : str
            The type of consumer. A single consumer, multiple fast consumers,
            multiple slow consumers. See ConsumerType constants.
        """
        self._topics = topics
        self._queue_name = queue
        self._parameters = parameters

        self._connection = pika.BlockingConnection(self.parameters)
        self._channel = self._connection.channel()
        self._channel.basic_qos(
            prefetch_count=ConsumerType.SPEED_TO_PREFETCH[consumer_type],
        )

        self._channel.queue_declare(
            queue=self._queue_name,
            durable=True,
        )

        self._bind_queue()

    @property
    def parameters(self):
        if self._parameters is None:
            self._parameters = RabbitConnectionParameters()
        return self._parameters.connection_parameters

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._channel.stop_consuming()
        self._connection.close()

    def consume_messages(self, consumer_callback):
        def callback_wrapper(ch, method, properties, body):
            correlation_id = '{}.{}'.format(
                properties.correlation_id,
                self._queue_name,
            )
            logging.basicConfig(
                format='[%(levelname)s] {correlation_id}: %(message)s'.format(
                    correlation_id=correlation_id
                ),
                level=logging.INFO,
            )
            consumer_callback(
                correlation_id,
                method.routing_key,
                bson.loads(body),
            )
            ch.basic_ack(delivery_tag=method.delivery_tag)

        self._channel.basic_consume(
            callback_wrapper,
            queue=self._queue_name,
        )
        self._channel.start_consuming()

    def _bind_queue(self):
        for topic in self._topics:
            exchange = Producer.get_exchange_name(topic)

            self._channel.exchange_declare(
                exchange=exchange,
                exchange_type='topic',
                durable=True,
            )
            self._channel.queue_bind(
                exchange=exchange,
                queue=self._queue_name,
                routing_key=topic,
            )



import bson

from franz import FranzEvent, InvalidMessage, SerializationError


class BaseProducer:
    def send_message(self, topic, message):
        self._check_message_is_valid_event(message)
        self._send_message_to_topic(topic, message)

    def _send_message_to_topic(self, topic, message):
        raise NotImplementedError(
            "_send_message_to_topic should be overridden in subclasses."
        )

    @staticmethod
    def serialize_message(message):
        """
        Serializes an object. It must subclass :class:`FranzEvent`.

        Parameters
        ----------
        message : FranzEvent
            The object to be serialized.

        Returns
        -------
        bytes
        """
        try:
            return bson.dumps(message.serialize())
        except bson.UnknownSerializerError:
            raise SerializationError(
                "Unable to serialize message: {}.".format(message)
            )

    @staticmethod
    def _check_message_is_valid_event(message):
        if not isinstance(message, FranzEvent):
            raise InvalidMessage(
                "{} is not a valid message. "
                "Must subclass `franz.FranzEvent`.".format(
                    type(message),
                )
            )



import pika


class RabbitConnectionParameters:
    _DEFAULT = pika.ConnectionParameters._DEFAULT

    def __init__(
        self, host='localhost', port=5672, virtual_host='/',
        username=None, password=None,
    ):
        self.host = host
        self.port = port
        self.virtual_host = virtual_host
        self.username = username
        self.password = password

    @property
    def connection_parameters(self):
        return pika.ConnectionParameters(
            host=self.host,
            port=self.port,
            credentials=self.credentials,
            virtual_host=self.virtual_host,
        )

    @property
    def credentials(self):
        if self.username and self.password:
            return pika.PlainCredentials(
                username=self.username,
                password=self.password,
            )
        return self._DEFAULT
